//
// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2020 - Stephane MOTTELET
//
// This file is hereby licensed under the terms of the GNU GPL v2.0,
// For more information, see the COPYING file which you should have received
//
//--------------------------------------------------------------------------

#include "function.hxx"
#include "double.hxx"
#include "sparse.hxx"
#include <Eigen/Sparse>
#include "string.hxx"
#include "struct.hxx"
#include "mlist.hxx"
#include "pointer.hxx"
#include "configvariable.hxx"
#include "commentexp.hxx"

#include "ColPackHeaders.h"

extern "C"
{
#include "localization.h"
#include "Scierror.h"
#include "sciprint.h"
}

types::Function::ReturnValue sci_numspjacobian(types::typed_list &in, int _iRetCount, types::typed_list &out)
{
    // H = numspjacobian(f,x0,sp_pattern);
    // f is a callable returning the function
    types::Callable *pCallFunction;
    char *pCallFunctionName;
    char errorMsg[256];
    types::typed_list pParameters;
    types::Double *pDblX0;
    types::Double *pDblSeed;
    types::Double *pDblColors;
    types::typed_list in2;
    types::typed_list out2;
    types::optional_list opt2;

    int iNbRows;
    int iNbCols;
    int iNonZeros;
    int *piValueColIndex;
    int *piRowBeginIndex;
    unsigned int **ppuiJacobianSparsityPattern;

    if (in.size()  != 3)
    {
        Scierror(999, _("%s: Wrong number of input argument(s): %d expected.\n"), "numspjacobian", 3);
        return types::Function::Error;
    }

    if (_iRetCount > 3)
    {
        Scierror(999, _("%s: Wrong number of output argument(s): at most %d expected.\n"), "numspjacobian", 3);
        return types::Function::Error;
    }

    if (in[0]->isCallable())
    {
        pCallFunction = in[0]->getAs<types::Callable>();
    }
    else if (in[0]->isList())
    {
        types::List *pList = in[0]->getAs<types::List>();
        types::InternalType *pIFirst =  pList->get(0);
        if (pIFirst->isCallable())
        {
            pCallFunction = pIFirst->getAs<types::Callable>();
        }
        else
        {
            Scierror(999,_("%s: first element of parameter %d should be a function.\n"), "numspjacobian", 1);
            return types::Function::Error;
        }
        for (int i=1; i<pList->getSize(); i++)
        {
            pParameters.push_back(pList->get(i));
        }
    }
    else
    {
        Scierror(999, _("%s: parameter %d should be a function or a list.\n"), "numspjacobian", 1);
        return types::Function::Error;
    }

    if (in[1]->isDouble() && in[1]->getAs<types::Double>()->isComplex() == false)
    {
        pDblX0 = in[1]->getAs<types::Double>();
    }
    else
    {
        Scierror(999, _("%s: Wrong type for input argument #%d : A real matrix expected.\n"), "numspjacobian", 2);
        return types::Function::Error;
    }

    if (in[2]->isSparse())
    {
        types::Sparse *pSpPattern = in[2]->getAs<types::Sparse>();
        types::Sparse::RealSparse_t* psp = pSpPattern->matrixReal;
        iNbRows = pSpPattern->getRows();
        iNbCols = pSpPattern->getCols();
        iNonZeros = pSpPattern->nonZeros();
        piValueColIndex = psp->innerIndexPtr();// column index of values
        piRowBeginIndex =  psp->outerIndexPtr(); // index of each row begining in values vector
    }
    else if (in[2]->isSparseBool())
    {
        types::SparseBool *pSpBoolPattern = in[2]->getAs<types::SparseBool>();
        types::SparseBool::BoolSparse_t* pspb = pSpBoolPattern->matrixBool;
        iNbRows = pSpBoolPattern->getRows();
        iNbCols = pSpBoolPattern->getCols();
        iNonZeros = pSpBoolPattern->nbTrue();
        piValueColIndex = pspb->innerIndexPtr();// column index of values
        piRowBeginIndex =  pspb->outerIndexPtr(); // index of each row begining in values vector
    }
    else
    {
        Scierror(999, _("%s: Wrong type for input argument #%d : A sparse matrix expected.\n"), "numspjacobian", 3);
        return types::Function::Error;
    }

    // Convert sparsity pattern of user matrix to internal ColPack compressed row format:
    ppuiJacobianSparsityPattern = new unsigned int*[iNbRows];
    int iValueIndex = 0;
    int iNonZerosInRow = 0;
    for (int i=1; i<iNbRows+1; i++)
    {
        iNonZerosInRow =  piRowBeginIndex[i]-piRowBeginIndex[i-1]; // we assume that matrix is in Eigen compressed row format
        ppuiJacobianSparsityPattern[i-1] = new unsigned int[iNonZerosInRow + 1];
        ppuiJacobianSparsityPattern[i-1][0] = iNonZerosInRow; // 0-index term holds the number of terms in row
        for (int j = 0; j < iNonZerosInRow; j++)
        {
            ppuiJacobianSparsityPattern[i-1][j+1] = piValueColIndex[iValueIndex++];
        }
    }

    //Step 2: Obtain the seed matrix via coloring.
    double*** pppdblSeed = new double**;
    int iSeedRowCount;
    int iSeedColumnCount;
    int iColorCount; //The number of distinct colors used to color the graph

    //Step 2.1: Read the sparsity pattern of the given Jacobian matrix (compressed sparse rows format)
    //and create the corresponding graph
    ColPack::BipartiteGraphPartialColoringInterface *g = new ColPack::BipartiteGraphPartialColoringInterface(SRC_MEM_ADOLC, ppuiJacobianSparsityPattern, iNbRows, iNbCols);

    //Step 2.2: Do Partial-Distance-Two-Coloring the bipartite graph with the specified ordering
    g->PartialDistanceTwoColoring( "SMALLEST_LAST", "COLUMN_PARTIAL_DISTANCE_TWO");
    //Step 2.3: From the coloring information, create and return the seed matrix
    (*pppdblSeed) = g->GetSeedMatrix(&iSeedRowCount, &iSeedColumnCount);

    // iNbSeedRows is the number of vertices of the graph, i.e. the number of Hessian columns
    //int iNbSeedRows = g->GetRowVertexCount();
    // iSeedColumnCount can be larger that the actual number of colors
    int iNbSeedCols = g->GetVertexColorCount();

    //Estimation of directional derivative of the gradient for
    //each column of seed matrix

    types::Double *pDblX = new types::Double(pDblX0->getSize(),2*iNbSeedCols);

    //prepare X for evaluation of gradient at X0±step*Seed(:,j) for each column j of Seed
    double dblStep = std::pow(2.0,-16.0);
    for (int i = 0; i < iNbRows; i++)
    {
        double dblX0i = pDblX0->get(i);
        for (int j = 0; j < iNbSeedCols; j++)
        {
            double dblH =  dblStep*(double)(*pppdblSeed)[i][j];
            pDblX->set(i,2*j,dblX0i-dblH);
            pDblX->set(i,2*j+1,dblX0i+dblH);
        }
    }

    in2.push_back(pDblX);
    pDblX->IncreaseRef();

    pCallFunctionName = wide_string_to_UTF8(pCallFunction->getName().c_str());
    try
    {
        // new std::wstring(L"") is deleted in destructor of ast::CommentExp
        pCallFunction->invoke(in2, opt2, 1, out2, ast::CommentExp(Location(), new std::wstring(L"")));
    }
    catch(ast::InternalError& ie)
    {
        if (ConfigVariable::getLastErrorFunction() == L"")
        {
            // necessary for errors such as "Wrong number of input arguments"
            std::wostringstream ostr;
            ostr << pCallFunctionName << L": " <<  ie.GetErrorMessage();
            ie.SetErrorMessage(ostr.str());
        }
        throw(ie);
    }

    pDblX->DecreaseRef();
    if (pDblX->isDeletable())
    {
        delete pDblX;
    }

    if (out2.size() != 1)
    {
        sprintf(errorMsg, _("%s: Wrong number of output argument(s): %d expected.\n"), pCallFunctionName, 1);
        throw ast::InternalError(errorMsg);
    }

    if (out2[0]->isDouble() == false)
    {
        sprintf(errorMsg, _("%s: Wrong type for output argument #%d: real matrix expected.\n"), pCallFunctionName,1);
        throw ast::InternalError(errorMsg);
    }

    types::Double *pDblFstep = out2[0]->getAs<types::Double>();

    if (pDblFstep->getRows() != iNbRows || pDblFstep->getCols() != 2*iNbSeedCols)
    {
        sprintf(errorMsg, _("%s: Wrong size for output argument #%d: a matrix with %d rows and %d columns expected.\n"), pCallFunctionName, 1, pDblX->getRows(), pDblX->getCols());
        throw ast::InternalError(errorMsg);
    }

    double **ppdblProd = new double*[iNbRows];
    types::Double *pDblProd = new types::Double(iNbRows,iNbSeedCols);
    pDblSeed = new types::Double(iNbRows,iNbSeedCols);

    for (int i = 0; i < iNbRows; i++)
    {
        ppdblProd[i] = new double[iNbSeedCols];
        for (int j = 0; j < iNbSeedCols; j++)
        {
            //centered/order 2 approximation of derivative
            ppdblProd[i][j] = (pDblFstep->get(i,2*j+1)-pDblFstep->get(i,2*j))/2.0/dblStep;
            pDblProd->set(i,j,(pDblFstep->get(i,2*j+1)-pDblFstep->get(i,2*j))/2.0/dblStep);
            pDblSeed->set(i,j,(*pppdblSeed)[i][j]);
        }
    }

    pDblFstep->killMe();

    //Step 4: Recover the numerical values of the original matrix from the compressed representation.
    // The new values are store in "ppHJacobianValue"
    double** ppJacobianValue;
    ColPack::JacobianRecovery1D* jr1d = new ColPack::JacobianRecovery1D();

    jr1d->RecoverD2Cln_RowCompressedFormat(g, ppdblProd, ppuiJacobianSparsityPattern, &ppJacobianValue);

    //convert to Eigen CSR: outer and inner vectors are already known from initial pattern.
    //hence we just concatenate the Hessian values as a single vector.
    double *pdblValues = new double[iNonZeros];

    iValueIndex = 0;
    for (int i = 0; i < iNbRows; i++)
    {
        for (int j = 1; j < (int)ppJacobianValue[i][0]+1; j++)
        {
            pdblValues[iValueIndex++] = ppJacobianValue[i][j];
        }
    }

    types::Sparse *pDblJacobian = new types::Sparse(iNbRows,iNbCols,iNonZeros,piValueColIndex,piRowBeginIndex,pdblValues,NULL);

    delete[] pdblValues;

    // ppJacobianValue deleted by destructor of hr
    // free_2DMatrix(ppJacobianValue,iNbRows);
    delete jr1d;

    // pppdblSeed deleted by destructor of g
    // free_2DMatrix(pppdblSeed,iNbRows);
    delete g;

    // user managed, hence to be explicitely deleted
    free_2DMatrix(ppdblProd,iNbRows);
    free_2DMatrix(ppuiJacobianSparsityPattern, iNbRows);

    out.push_back(pDblJacobian);
    out.push_back(pDblSeed);
    out.push_back(pDblProd);


    return types::Function::OK;
}

